#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mysql/mysql.h>
#include <errno.h>

#include "erl_driver.h"
#ifndef ERL_DRV_NIL
#include "erl_driver_compat.h"
#endif

/* op codes */
#define ID_CONNECT	0
#define ID_QUERY	1
#define ID_RESULTINFO	2
#define ID_FETCHRESULT	3
#define ID_ESCAPESTRING 4
#define ID_SETCHARSET	5

#define ASIZE(spec)	sizeof(spec)/sizeof(spec[0])

/* atoms which are sent to erlang */
static ErlDrvTermData am_ok;
static ErlDrvTermData am_error;
static ErlDrvTermData am_einval;
static ErlDrvTermData am_eof;
static ErlDrvTermData am_row;
static ErlDrvTermData am_null;

static ErlDrvEntry mysql_drv_entry;

typedef struct mysqldrv_t {
	ErlDrvPort	port;
	ErlDrvTermData	port_id;          /* the port identifier as ErlDrvTermData */
	ErlDrvPDL	lock;
	ErlDrvMutex	*mtx;
	ErlDrvCond	*cv;
	struct task	*head;
	struct task	*tail;
	ErlDrvTid	tid;
	int		shutdown;
	MYSQL		*mysql;
	MYSQL_RES	*res;
} mysqldrv_t;

struct task {
	struct task	*next;
	int		type;
};

struct task_connect {
	struct task	*next;
	int		type;
	char		*host;
	char		*user;
	char		*pass;
	char		*db;
	char		*unixsock;
	short		port_num;
	int		flags;
};

struct task_query {
	struct task	*next;
	int		type;
	char		*str;
};

struct task_escape_string {
	struct task	*next;
	int		type;
	char		*str;
};

struct task_set_charset {
	struct task	*next;
	int		type;
	char		*str;
};

struct termdata {
	ErlDrvTermData *spec;
	int		len;
	int		size;
};

static void
init_term_data(struct termdata *termdat, int siz)
{
	termdat->spec = malloc(siz * sizeof(ErlDrvTermData));
	termdat->size = siz;
	termdat->len  = 0;
}

static void
free_term_data(struct termdata *termdat)
{
	free(termdat->spec);
}

static void
add_term_data(struct termdata *termdat, ErlDrvTermData dat)
{
	if (termdat->len >= termdat->size) {
		int newsize = termdat->size * 2;
		ErlDrvTermData *newspec = realloc(termdat->spec, newsize * sizeof(ErlDrvTermData));
		if (newspec == NULL) {
			fprintf(stderr, "not enough memory\n");
			abort();
		}
		termdat->spec = newspec;
		termdat->size = newsize;
	}
	termdat->spec[termdat->len] = dat;
	termdat->len++;
}

static void
reset_term_data(struct termdata *termdat)
{
	termdat->len = 0;
}

static void
add_term_data2(struct termdata *termdat, ErlDrvTermData dat1, ErlDrvTermData dat2)
{
	add_term_data(termdat, dat1);
	add_term_data(termdat, dat2);
}

static void
add_term_data3(struct termdata *termdat, ErlDrvTermData dat1, ErlDrvTermData dat2, ErlDrvTermData dat3)
{
	add_term_data(termdat, dat1);
	add_term_data(termdat, dat2);
	add_term_data(termdat, dat3);
}

static void
add_term_data4(struct termdata *termdat, ErlDrvTermData dat1, ErlDrvTermData dat2, ErlDrvTermData dat3,
		ErlDrvTermData dat4)
{
	add_term_data(termdat, dat1);
	add_term_data(termdat, dat2);
	add_term_data(termdat, dat3);
	add_term_data(termdat, dat4);
}

static void *task_loop(void *arg);
static void free_task(struct task *t);

static ErlDrvData
mysqldrv_start(ErlDrvPort port, char *buf)
{
	mysqldrv_t *dd = (mysqldrv_t *) driver_alloc(sizeof(mysqldrv_t));

	if (dd == NULL)
		return (ErlDrvData) -1;
	dd->port = port;
	dd->port_id = driver_mk_port(port);
	dd->mysql = mysql_init(NULL);
	dd->head = NULL;
	dd->tail = NULL;
	dd->lock = driver_pdl_create(dd->port);
	dd->mtx = erl_drv_mutex_create("mysql_drv queue mtx");
	dd->cv = erl_drv_cond_create("mysql_drv queue cond");
	dd->shutdown = 0;
	dd->res = NULL;
	erl_drv_thread_create("mysql_drv thread", &dd->tid, task_loop, dd, NULL);
	return (ErlDrvData) dd;
}

static void
mysqldrv_stop(ErlDrvData drv_data)
{
	mysqldrv_t *dd = (mysqldrv_t *) drv_data;
	struct task *t;

	erl_drv_mutex_lock(dd->mtx);
	dd->shutdown = 1;
	erl_drv_cond_broadcast(dd->cv);
	erl_drv_mutex_unlock(dd->mtx);
	erl_drv_thread_join(dd->tid, NULL);
	t = dd->head;
	while (t) {
		struct task *t1 = t->next;
		free_task(t);
		t = t1;
	}
	if (dd->res)
		mysql_free_result(dd->res);
	mysql_close(dd->mysql);
	erl_drv_mutex_destroy(dd->mtx);
	erl_drv_cond_destroy(dd->cv);
	driver_free(dd);
}

/* send {P, ok} to caller */
static int
driver_send_ok(mysqldrv_t *dd)
{
	ErlDrvTermData spec[] = {
		ERL_DRV_PORT, dd->port_id,
		ERL_DRV_ATOM, am_ok,
		ERL_DRV_TUPLE, 2};

	return driver_send_term(dd->port, driver_connected(dd->port), spec, ASIZE(spec));
}

/* send {P, error, Error} to caller */
static int
driver_send_error(mysqldrv_t *dd, ErlDrvTermData *am)
{
	ErlDrvTermData spec[] = {
		ERL_DRV_PORT, dd->port_id,
		ERL_DRV_ATOM, am_error,
		ERL_DRV_ATOM, *am,
		ERL_DRV_TUPLE, 3 };
	return driver_send_term(dd->port, driver_connected(dd->port), spec, ASIZE(spec));
}

static int
get_int16(char **buf, char *end, short *val)
{
	if (*buf +2 > end)
		return (-1);

	*val = ((unsigned char *)(*buf))[0] << 8 |
	       ((unsigned char *)(*buf))[1];
	*buf += 2;
	return (0);
}

static int
get_int32(char **buf, char *end, int *val)
{
	if (*buf + 4 > end)
		return (-1);

	*val = ((unsigned char *)(*buf))[0] << 24 | \
	       ((unsigned char *)(*buf))[1] << 16 | \
	       ((unsigned char *)(*buf))[2] << 8  | \
	       ((unsigned char *)(*buf))[3];
	*buf += 4;
	return (0);
}

static int
get_string(char **buf, char *end, char **string)
{
	int len;

	*string = NULL;
	if (get_int32(buf, end, &len) == -1)
		return (-1);
	if (end - *buf < len)
		return (-1);
	*string = strndup(*buf, len);
	*buf += len;
	return (0);
}

static void
send_mysql_error(mysqldrv_t *dd, int success)
{
	if (!success) {
		const char *err_str = mysql_error(dd->mysql);
		ErlDrvTermData spec[] = {
			ERL_DRV_PORT, dd->port_id,
			ERL_DRV_ATOM, am_error,
			ERL_DRV_STRING, (ErlDrvTermData)err_str, strlen(err_str),
			ERL_DRV_TUPLE, 3};

		driver_pdl_lock(dd->lock);
		driver_send_term(dd->port, driver_connected(dd->port), spec, ASIZE(spec));
		driver_pdl_unlock(dd->lock);
	} else {
		driver_pdl_lock(dd->lock);
		driver_send_ok(dd);
		driver_pdl_unlock(dd->lock);
	}
}

static void
mysqldrv_connect(mysqldrv_t *dd, const char *host, const char*user, const char *pass,
	const char *db, short port_num, const char *unixsock, int flags)
{
	if (!mysql_real_connect(dd->mysql,
		host[0] ? host : NULL,
		user, 
		pass,
		db[0] ? db : NULL,
		port_num,
		unixsock[0] ? unixsock : NULL,
		flags)) {

		send_mysql_error(dd, 0);
	} else {
		send_mysql_error(dd, 1);
	}
}

static void
mysqldrv_query(mysqldrv_t *dd, const char *str)
{
	if (dd->res) {
		mysql_free_result(dd->res);
		dd->res = NULL;
	}

	if (mysql_query(dd->mysql, str)) {
		send_mysql_error(dd, 0);
		return;
	}

	dd->res = mysql_use_result(dd->mysql);
	if (dd->res == NULL)
		return (send_mysql_error(dd, mysql_field_count(dd->mysql) == 0));

	send_mysql_error(dd, 1);
}

static void
mysqldrv_result_info(mysqldrv_t *dd)
{
	unsigned int i, rows;
	MYSQL_FIELD *field;
	struct termdata termdata;

	int affected_rows = mysql_affected_rows(dd->mysql);
	if (dd->res == NULL) {
		ErlDrvTermData spec[] = {
			ERL_DRV_PORT, dd->port_id,
			ERL_DRV_ATOM, am_ok,
			ERL_DRV_NIL, 
			ERL_DRV_LIST,1,
			ERL_DRV_INT, affected_rows,
			ERL_DRV_TUPLE,4
			};
		driver_pdl_lock(dd->lock);
		driver_send_term(dd->port, driver_connected(dd->port), spec, ASIZE(spec));
		driver_pdl_unlock(dd->lock);
		return;
	}

	init_term_data(&termdata, 100);
	rows = mysql_num_fields(dd->res);
	add_term_data2(&termdata, ERL_DRV_PORT, dd->port_id);
	add_term_data2(&termdata, ERL_DRV_ATOM, am_ok);
	for (i = 0; i < rows; ++i) {
		field = mysql_fetch_field_direct(dd->res, i);

		add_term_data3(&termdata, ERL_DRV_STRING, (ErlDrvTermData)field->name, field->name_length);
		add_term_data3(&termdata, ERL_DRV_STRING, (ErlDrvTermData)field->org_name, field->org_name_length);
		add_term_data3(&termdata, ERL_DRV_STRING, (ErlDrvTermData)field->table, field->table_length);
		add_term_data3(&termdata, ERL_DRV_STRING, (ErlDrvTermData)field->org_table, field->org_table_length);
		add_term_data3(&termdata, ERL_DRV_STRING, (ErlDrvTermData)field->db, field->db_length);
		add_term_data3(&termdata, ERL_DRV_STRING, (ErlDrvTermData)field->catalog, field->catalog_length);
		add_term_data3(&termdata, ERL_DRV_STRING, (ErlDrvTermData)(field->def ? field->def : ""), field->def_length);
		add_term_data2(&termdata, ERL_DRV_INT, field->length);
		add_term_data2(&termdata, ERL_DRV_INT, field->max_length);
		add_term_data2(&termdata, ERL_DRV_INT, field->flags);
		add_term_data2(&termdata, ERL_DRV_INT, field->decimals);
		add_term_data2(&termdata, ERL_DRV_INT, field->charsetnr);
		add_term_data2(&termdata, ERL_DRV_INT, field->type);
		add_term_data2(&termdata, ERL_DRV_TUPLE, 13);
	}

	add_term_data(&termdata, ERL_DRV_NIL);
	add_term_data2(&termdata, ERL_DRV_LIST, rows+1);
	add_term_data2(&termdata, ERL_DRV_INT, affected_rows);
	add_term_data2(&termdata, ERL_DRV_TUPLE, 4);

	driver_pdl_lock(dd->lock);
	driver_send_term(dd->port, driver_connected(dd->port), termdata.spec, termdata.len);
	driver_pdl_unlock(dd->lock);
	free_term_data(&termdata);
}

static void
mysqldrv_fetch_result(mysqldrv_t *dd)
{
	if (dd->res) {
		struct termdata termdata;
		MYSQL_ROW row;
		int num_fields = mysql_num_fields(dd->res);

		init_term_data(&termdata, (num_fields + 4) * 3);
		while((row = mysql_fetch_row(dd->res))) {
			reset_term_data(&termdata);

			add_term_data2(&termdata, ERL_DRV_PORT, dd->port_id);
			add_term_data2(&termdata, ERL_DRV_ATOM, am_row);

			int i;
			for (i = 0; i < num_fields; ++i) {
				if (row[i] != NULL)
					add_term_data3(&termdata, ERL_DRV_STRING,
						(ErlDrvTermData)row[i], strlen(row[i]));
				else 
					add_term_data2(&termdata, ERL_DRV_ATOM, am_null);
			}

			add_term_data(&termdata, ERL_DRV_NIL);
			add_term_data2(&termdata, ERL_DRV_LIST, num_fields+1);

			add_term_data2(&termdata, ERL_DRV_TUPLE, 3);

			driver_pdl_lock(dd->lock);
			driver_send_term(dd->port, driver_connected(dd->port), termdata.spec, termdata.len);
			driver_pdl_unlock(dd->lock);
		}

		free_term_data(&termdata);
	}
	ErlDrvTermData spec[] = {
		ERL_DRV_PORT, dd->port_id,
		ERL_DRV_ATOM, am_eof,
		ERL_DRV_TUPLE,2
	};
	driver_pdl_lock(dd->lock);
	driver_send_term(dd->port, driver_connected(dd->port), spec, ASIZE(spec));
	driver_pdl_unlock(dd->lock);

	if (dd->res) {
		mysql_free_result(dd->res);
		dd->res = NULL;
	}
	return;
}

static void
mysqldrv_escape_string(mysqldrv_t *dd, const char *str)
{
	int len = strlen(str);
	char *to;

	to = malloc(len * 2 + 1);
	mysql_real_escape_string(dd->mysql, to, str, len);

	ErlDrvTermData spec[] = {
		ERL_DRV_PORT, dd->port_id,
		ERL_DRV_ATOM, am_ok,
		ERL_DRV_STRING, (ErlDrvTermData)to, len,
		ERL_DRV_TUPLE, 3
	};

	driver_pdl_lock(dd->lock);
	driver_send_term(dd->port, driver_connected(dd->port), spec, ASIZE(spec));
	driver_pdl_unlock(dd->lock);
}

static void
mysqldrv_set_charset(mysqldrv_t *dd, const char *name)
{
	int ret = mysql_set_character_set(dd->mysql, name);
	send_mysql_error(dd, ret == 0);
}

static struct task *
deq(mysqldrv_t *dd)
{
	struct task *t;

	if (dd->head == NULL)
		return (NULL);

	t = dd->head;
	dd->head = t->next;
	if (t == dd->tail)
		dd->tail = dd->head = NULL;
	return (t);
}

static void
enq(mysqldrv_t *dd, struct task *t)
{
	t->next = NULL;
	if (dd->head == NULL) {
		dd->head = dd->tail = t;
	} else {
		dd->tail->next = t;
		dd->tail = t;
	}
}

static void
enq_task(mysqldrv_t *dd, struct task *t)
{
	erl_drv_mutex_lock(dd->mtx);
	enq(dd, t);
	erl_drv_cond_signal(dd->cv);
	erl_drv_mutex_unlock(dd->mtx);
}

static void *
task_loop(void *arg)
{
	mysqldrv_t *dd = arg;
	struct task *t;

loop:
	erl_drv_mutex_lock(dd->mtx);
	while (dd->head == NULL && dd->shutdown == 0) {
		erl_drv_cond_wait(dd->cv, dd->mtx);
	}
	if (dd->shutdown) {
		erl_drv_mutex_unlock(dd->mtx);
		return NULL;
	}
	t = deq(dd);
	erl_drv_mutex_unlock(dd->mtx);

	switch(t->type) {
	case ID_CONNECT: {
		struct task_connect *t1 = (struct task_connect *)t;
		mysqldrv_connect(dd, t1->host, t1->user, t1->pass, t1->db, 
			t1->port_num, t1->unixsock, t1->flags);
		free(t1->host);
		free(t1->user);
		free(t1->pass);
		free(t1->db);
		free(t1->unixsock);
		free(t1);
		}
		break;

	case ID_QUERY: {
		struct task_query *t1 = (struct task_query *)t;
		mysqldrv_query(dd, t1->str);
		free(t1->str);
		free(t1);
		}
		break;

	case ID_RESULTINFO: {
		mysqldrv_result_info(dd);
		free(t);
		}
		break;

	case ID_FETCHRESULT: {
		mysqldrv_fetch_result(dd);
		free(t);
		}
		break;

	case ID_ESCAPESTRING: {
		struct task_escape_string *t1 = (struct task_escape_string *)t;
		mysqldrv_escape_string(dd, t1->str);
		free(t1->str);
		free(t1);
		}
		break;

	case ID_SETCHARSET: {
		struct task_set_charset *t1 = (struct task_set_charset *)t;
		mysqldrv_set_charset(dd, t1->str);
		free(t1->str);
		free(t1);
		}
		break;
	default:
		fprintf(stderr, "mysql_drv:%s unknown packet:%d\n", __func__, t->type);
		free(t);
	}

	goto loop;
}

static void
free_task(struct task *t)
{
	switch(t->type) {
	case ID_CONNECT: {
		struct task_connect *t1 = (struct task_connect *)t;
		free(t1->host);
		free(t1->user);
		free(t1->pass);
		free(t1->db);
		free(t1->unixsock);
		free(t1);
		}
		break;

	case ID_QUERY: {
		struct task_query *t1 = (struct task_query *)t;
		free(t1->str);
		free(t1);
		}
		break;

	case ID_RESULTINFO: {
		free(t);
		}
		break;

	case ID_FETCHRESULT: {
		free(t);
		}
		break;

	case ID_ESCAPESTRING: {
		struct task_escape_string *t1 = (struct task_escape_string *)t;
		free(t1->str);
		free(t1);
		}
		break;

	case ID_SETCHARSET: {
		struct task_set_charset *t1 = (struct task_set_charset *)t;
		free(t1->str);
		free(t1);
		}
		break;
	default:
		fprintf(stderr, "mysql_drv:%s unknown packet:%d\n", __func__, t->type);
		free(t);
	}
}

static void
mysqldrv_from_erlang(ErlDrvData drv_data, char *buf, int len)
{
	mysqldrv_t *dd = (mysqldrv_t *)drv_data;
	char *end, *bp = buf;
	unsigned int i;

	i = bp[0];
	end = bp + len;
	bp++;

	switch (i) {
	case ID_CONNECT: {
		char *host=NULL, *user=NULL, *pass=NULL, *db=NULL, *unixsock=NULL;
		short port_num;
		int flags;
		if (get_string(&bp, end, &host) == -1 ||
		    get_string(&bp, end, &user) == -1 ||
		    get_string(&bp, end, &pass) == -1 ||
		    get_string(&bp, end, &db) == -1 ||
		    get_int16(&bp, end, &port_num) == -1 ||
		    get_string(&bp, end, &unixsock) == -1||
		    get_int32(&bp, end, &flags) == -1) {
			free(host);
			free(user);
			free(pass);
			free(db);
			free(unixsock);
			driver_send_error(dd, &am_einval);
			return;
		}
		struct task_connect *t = malloc(sizeof(struct task_connect));

		t->type = ID_CONNECT;
		t->host = host;
		t->user = user;
		t->pass = pass;
		t->db = db;
		t->port_num = port_num;
		t->unixsock = unixsock;
		t->flags = flags;
		
		enq_task(dd, (struct task *)t);
		}
		break;

	case ID_QUERY: {
		char *str;

		if (get_string(&bp, end, &str) == -1) {
			driver_send_error(dd, &am_einval);
			return;
		}

		struct task_query *t = malloc(sizeof(struct task_query));
		t->type = ID_QUERY;
		t->str  = str;
		enq_task(dd, (struct task *)t);
		}
		break;

	case ID_RESULTINFO: {
		struct task *t = malloc(sizeof(struct task));
		t->type = ID_RESULTINFO;
		enq_task(dd, (struct task *)t);
		}
		break;

	case ID_FETCHRESULT: {
		struct task *t = malloc(sizeof(struct task));
		t->type = ID_FETCHRESULT;
		enq_task(dd, (struct task *)t);
		}
		break;

	case ID_ESCAPESTRING: {
		char *str;

		if (get_string(&bp, end, &str) == -1) {
			driver_send_error(dd, &am_einval);
			return;
		}
		struct task_escape_string *t = malloc(sizeof(struct task_escape_string));
		t->type = ID_ESCAPESTRING;
		t->str = str;
		enq_task(dd, (struct task *)t);
		}
		break;

	case ID_SETCHARSET: {
		char *str;
		if (get_string(&bp, end, &str) == -1) {
			driver_send_error(dd, &am_einval);
			return;
		}

		struct task_set_charset *t = malloc(sizeof(struct task_set_charset));
		t->type = ID_SETCHARSET;
		t->str = str;
		enq_task(dd, (struct task *)t);
		}
	} /* switch */
	return;
}

DRIVER_INIT(mysql_drv)
{
	am_ok           = driver_mk_atom("ok");
	am_error        = driver_mk_atom("error");
	am_einval       = driver_mk_atom("einval");
	am_eof		= driver_mk_atom("eof");
	am_row		= driver_mk_atom("row");
	am_null		= driver_mk_atom("null");

	mysql_drv_entry.init         = NULL;
	mysql_drv_entry.start        = mysqldrv_start;
	mysql_drv_entry.stop         = mysqldrv_stop;
	mysql_drv_entry.output       = mysqldrv_from_erlang;
	mysql_drv_entry.ready_input  = NULL;
	mysql_drv_entry.ready_output = NULL;
	mysql_drv_entry.driver_name  = "mysql_drv";
	mysql_drv_entry.finish       = NULL;
	mysql_drv_entry.outputv      = NULL;
	mysql_drv_entry.extended_marker = ERL_DRV_EXTENDED_MARKER;
	mysql_drv_entry.major_version = ERL_DRV_EXTENDED_MAJOR_VERSION;
	mysql_drv_entry.minor_version = ERL_DRV_EXTENDED_MINOR_VERSION;
	mysql_drv_entry.driver_flags = ERL_DRV_FLAG_USE_PORT_LOCKING;
	return &mysql_drv_entry;
}
