-module(mysql).
-export([init/0, open/0, close/1, connect/8, fetch/2, transaction/2, escape_string/2]).
-export([set_charset/2]).
-define(DRV_NAME, "mysql_drv").

-define(ID_CONNECT,0).
-define(ID_QUERY,1).
-define(ID_RESULTINFO,2).
-define(ID_FETCHRESULT,3).
-define(ID_ESCAPESTRING,4).
-define(ID_SETCHARSET,5).

-include("mysql.hrl").

init() ->
	erl_ddll:start(),
	Path = code:priv_dir(mysql),
	erl_ddll:load(Path, ?DRV_NAME).

open() ->
	{ok, open_port({spawn, ?DRV_NAME}, [binary])}.

connect(Conn, Host, User, Password, Database, SockPort, UnixSocket, Options) ->
	Host1 = list2bin(Host),
	Host1Size = size(Host1),
	User1 = list2bin(User),
	User1Size = size(User1),
	Pass1 = list2bin(Password),
	Pass1Size = size(Pass1),
	Database1 = list2bin(Database),
	Database1Size = size(Database1),
	UnixSocket1 = list2bin(UnixSocket),
	UnixSocket1Size = size(UnixSocket1),
	Msg = << ?ID_CONNECT:8,
		Host1Size:32/integer,
		Host1:Host1Size/binary,

		User1Size:32/integer,
		User1:User1Size/binary,

		Pass1Size:32/integer,
		Pass1:Pass1Size/binary,

		Database1Size:32/integer,
		Database1:Database1Size/binary,

		SockPort:16/integer,

		UnixSocket1Size:32/integer,
		UnixSocket1:UnixSocket1Size/binary,

		Options:32/integer
	      >>,
	port_command(Conn, [Msg]),
	receive
		{Conn, ok} ->	ok;
		{Conn, error, ErrMsg} -> {error, ErrMsg}
	end.

close(Conn) ->
	port_close(Conn),
	ok.

fetch(Conn, Str) ->
	Str1 = list2bin(Str),
	Str1Size = size(Str1),
	Msg = <<?ID_QUERY:8, Str1Size:32/integer, Str1:Str1Size/binary>>,
	port_command(Conn, [Msg]),
	receive
		{Conn, ok} -> get_result(Conn);
		{Conn, error, ErrMsg} -> {error, ErrMsg}
	end.

get_result(Conn) ->
	case get_resultinfo(Conn) of
		{error, ErrMsg} -> {error, ErrMsg};
		{ok, Info} ->
			Msg = <<?ID_FETCHRESULT:8>>,
			port_command(Conn, Msg),
			Result1 = lists:reverse(receive_result_set(Conn, [])),
			Result2 = lists:reverse(convert_result_set(Result1, Info#mysql_resultinfo.field_info, [])),
			{ok, #mysql_result{result_info=Info, rows=Result2}}
	end.

get_resultinfo(Conn) ->
	Msg = <<?ID_RESULTINFO:8>>,
	port_command(Conn, [Msg]),
	receive
		{Conn, ok, FieldInfo, AffectedRows} ->
			NewFieldInfo = lists:reverse(translate_fieldinfo(FieldInfo,[])),
			ResultInfo = #mysql_resultinfo{field_info=NewFieldInfo,
					affected_rows = AffectedRows},
			{ok, ResultInfo};
		{Conn, error, ErrMsg} -> {error, ErrMsg}
	end.

translate_fieldinfo([], Result) ->
	Result;
translate_fieldinfo([H|T], Result) ->
	{Name, OrgName, Table, OrgTable, Db, Catalog, Def, Length, MaxLength, Flags, Decimals, Charsetnr, Type} = H,
	X = #mysql_fieldinfo{
		name = Name,
		org_name = OrgName,
		table = Table,
		org_table = OrgTable,
		db = Db,
		catalog = Catalog,
		def = Def,
		length = Length,
		max_length = MaxLength,
		flags = Flags,
		decimals = Decimals,
		charsetnr = Charsetnr,
		type = Type},
	translate_fieldinfo(T, [X|Result]).
	
receive_result_set(Conn, Result) ->
	receive
		{Conn, eof} ->Result;
		{Conn, row, Row} ->receive_result_set(Conn, [Row|Result])
	end.

transaction(Conn, Fun)->
	case fetch(Conn, "START TRANSACTION") of
	{ok, _Result} ->
		try
			Val = Fun(),
			case fetch(Conn, "COMMIT") of
				{ok, _Result2} ->
					{atomic, Val};
				Other ->
					 {aborted, Other, {rollback_result, fetch(Conn, "ROLLBACK")}}
			end
		catch
			_:X -> 
				{aborted, X, {rollback_result, fetch(Conn, "ROLLBACK")}}
		end;
	Other ->
		{aborted, Other, {rollback_result, ok}}
	end.

convert_result_set([], _, Result) ->
	Result;
convert_result_set([R|T], Info, Result) ->
	R1 = lists:reverse(convert_record(R, Info, [])),
	convert_result_set(T, Info, [R1|Result]).

convert_record([], _, Result) ->
	Result;
convert_record([Val|T], [Info|InfoT], Result) ->
	NewVal = convert_field(Val, Info#mysql_fieldinfo.type),
	convert_record(T, InfoT, [NewVal|Result]).

convert_field(null, _) ->
	null;
convert_field(Val, Type) ->
	case Type of
		T when T == ?MYSQL_TYPE_TINY;
		       T == ?MYSQL_TYPE_SHORT;
		       T == ?MYSQL_TYPE_LONG;
		       T == ?MYSQL_TYPE_LONGLONG;
		       T == ?MYSQL_TYPE_INT24;
		       T == ?MYSQL_TYPE_YEAR ->
		    list_to_integer(Val);

		T when T == ?MYSQL_TYPE_TIMESTAMP;
		       T == ?MYSQL_TYPE_DATETIME ->
		    {ok, [Year, Month, Day, Hour, Minute, Second], _} =
		    io_lib:fread("~d-~d-~d ~d:~d:~d", Val),
		    {datetime, {{Year, Month, Day}, {Hour, Minute, Second}}};
		?MYSQL_TYPE_TIME ->
		    {ok, [Hour, Minute, Second], _} =
		    io_lib:fread("~d:~d:~d", Val),
		    {time, {Hour, Minute, Second}};
		?MYSQL_TYPE_DATE ->
		    {ok, [Year, Month, Day], _} =
		    io_lib:fread("~d-~d-~d", Val),
		    {date, {Year, Month, Day}};
		T when T == ?MYSQL_TYPE_DECIMAL;
		       T == ?MYSQL_TYPE_NEWDECIMAL;
		       T == ?MYSQL_TYPE_FLOAT;
		       T == ?MYSQL_TYPE_DOUBLE ->
		    {ok, [Num], _} =
		        case io_lib:fread("~f", Val) of
                    	    {error, _} ->
				io_lib:fread("~d", Val);
			     Res ->
				Res
			end,
		    Num;
		_Other ->
		    Val
	end.
escape_string(Conn, String) ->
	Bin = list2bin(String),
	BinSize = size(Bin),
	Msg = << ?ID_ESCAPESTRING:8,BinSize:32/integer,Bin/binary >>,
	port_command(Conn, Msg),
	receive
		{Conn, ok, String2} -> String2
	end.

set_charset(Conn, Name) ->
	Bin = list2bin(Name),
	BinSize = size(Bin),
	Msg = << ?ID_SETCHARSET:8, BinSize:32/integer,Bin/binary >>,
	port_command(Conn, Msg),
	receive
		{Conn, ok} -> ok;
		{Conn, error, ErrMsg} -> {error, ErrMsg}
	end.

list2bin(L) when is_list(L)   -> list_to_binary(L);
list2bin(B) when is_binary(B) -> B.
